"""Simulation of 3D volume and associated helper functions."""

from typing import Literal

import einops
import numpy as np
import torch
from torch_fourier_filter.ctf import calculate_relativistic_electron_wavelength
from torch_fourier_filter.dose_weight import cumulative_dose_filter_3d
from torch_fourier_filter.mtf import make_mtf_grid

from ttsim3d.grid_coords import (
    fourier_rescale_3d_force_size,
    get_atom_voxel_indices,
    get_upsampling,
    get_voxel_neighborhood_offsets,
)
from ttsim3d.scattering_potential import (
    get_scattering_potential_of_voxel_batch,
)

BOND_SCALING_FACTOR = 1.043
PIXEL_OFFSET = 0.5
MAX_SIZE = 1536
ALLOWED_DOSE_FILTER_MODIFICATIONS = ["None", "sqrt", "rel_diff"]


def _calculate_lead_term(beam_energy_kev: float, sim_pixel_spacing: float) -> float:
    """Calculate the lead term for the scattering potential.

    Parameters
    ----------
    beam_energy_kev : float
        The beam energy in kV.
    sim_pixel_spacing : float
        The pixel spacing for the final simulation in Angstroms.

    Returns
    -------
    float
        The lead term for the scattering potential.
    """
    beam_energy_ev = beam_energy_kev * 1000
    wavelength_m = calculate_relativistic_electron_wavelength(beam_energy_ev)
    wavelength_A = wavelength_m * 1e10
    lead_term = BOND_SCALING_FACTOR * wavelength_A / 8.0 / (sim_pixel_spacing**2)

    return float(lead_term)


def _validate_dose_filter_inputs(
    dose_filter_modify_signal: str,
    dose_filter_critical_bfactor: float,
) -> None:
    """Ensure dose filter inputs are valid."""
    if dose_filter_modify_signal not in ALLOWED_DOSE_FILTER_MODIFICATIONS:
        raise ValueError(
            f"Invalid dose filter modification method: {dose_filter_modify_signal}. "
            f"Allowed methods are: {ALLOWED_DOSE_FILTER_MODIFICATIONS}"
        )

    if dose_filter_critical_bfactor == -1:
        return

    if dose_filter_critical_bfactor < 0:
        raise ValueError(
            "Critical B factor for dose filter must either be -1 (use critical "
            "exposure dose weighting) or a positive float. Given: "
            f"{dose_filter_critical_bfactor}"
        )


def _validate_dqe_filter_inputs(
    apply_dqe: bool,
    mtf_frequencies: torch.Tensor,
    mtf_amplitudes: torch.Tensor,
) -> None:
    """Ensure DQE filter inputs are valid."""
    if apply_dqe:
        if mtf_frequencies is None:
            raise ValueError(
                "If 'apply_dqe' is True, 'mtf_frequencies' must be provided. "
                "Got 'None'."
            )

        if mtf_amplitudes is None:
            raise ValueError(
                "If 'apply_dqe' is True, 'mtf_amplitudes' must be provided. "
                "Got 'None'."
            )

        if mtf_amplitudes.shape != mtf_frequencies.shape:
            raise ValueError(
                "The 'mtf_frequencies' and 'mtf_amplitudes' tensors must have "
                "the same shape. Got shapes: "
                f"{mtf_frequencies.shape} and {mtf_amplitudes.shape}."
            )


def _setup_sim3d_upsampling(
    sim_pixel_spacing: float,
    sim_volume_shape: tuple[int, int, int],
    upsampling: int,
) -> tuple[int, float, tuple[int, int, int]]:
    """Helper function to calculate upsampling factor and related values.

    Parameters
    ----------
    sim_pixel_spacing : float
        Desired final pixel spacing for simulation in Angstroms.
    sim_volume_shape : tuple[int, int, int]
        Desired final shape of the simulation volume. NOTE: simulations currently
        only support cubic volumes, but this is not explicitly checked.
    upsampling : int
        The upsampling factor as an int greater than 1 or -1. If -1, the
        upsampling factor is calculated automatically.

    Returns
    -------
    tuple[int, float, tuple[int, int, int]]
        The upsampling factor, the upsampled pixel size, and the upsampled shape.
    """
    if upsampling == -1:
        upsampling = get_upsampling(
            sim_pixel_spacing, sim_volume_shape[0], max_size=MAX_SIZE
        )
    elif upsampling < 1:
        raise ValueError("Upsampling factor must be greater than 1 (or -1 for auto)")

    upsampled_pixel_size = sim_pixel_spacing / upsampling
    upsampled_shape = tuple(np.array(sim_volume_shape) * upsampling)

    return upsampling, upsampled_pixel_size, upsampled_shape


def _setup_upsampling_coords(
    atom_positions_zyx: torch.Tensor,  # shape (N, 3)
    upsampled_pixel_size: float,
    upsampled_shape: tuple[int, int, int],
    mean_b_factor: float,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Helper function to calculate the voxel indices and offsets for upsampling.

    Parameters
    ----------
    atom_positions_zyx : torch.Tensor
        The atom coordinates (continuous) in Angstroms. Shape (N, 3) where N is
        the number of atoms.
    upsampled_pixel_size : float
        The pixel size in Angstroms for the upsampled volume.
    upsampled_shape : tuple[int, int, int]
        The shape of the upsampled volume.
    mean_b_factor : float
        The mean B factor of the atoms. Used to estimate the neighborhood size
        necessary for simulation.

    Returns
    -------
    tuple[torch.Tensor, torch.Tensor, torch.Tensor]
        The atom indices, atom dds, and voxel offsets for the neighborhood.
        Each tensor has shape (N, 3).
    """
    # Calculate the voxel coordinates of each atom
    atom_indices, atom_dds = get_atom_voxel_indices(
        atom_zyx=atom_positions_zyx,
        upsampled_pixel_size=upsampled_pixel_size,
        upsampled_shape=upsampled_shape,
        offset=PIXEL_OFFSET,
    )

    # Get the voxel offsets for the neighborhood around the atom voxel
    voxel_offsets_flat = get_voxel_neighborhood_offsets(
        mean_b_factor=mean_b_factor,
        upsampled_pixel_size=upsampled_pixel_size,
    )

    return atom_indices, atom_dds, voxel_offsets_flat


def simulate_atomwise_scattering_potentials(
    atom_positions_zyx: torch.Tensor,  # shape (N, 3)
    atom_ids: list[str],
    atom_b_factors: torch.Tensor,
    sim_pixel_spacing: float,
    sim_volume_shape: tuple[int, int, int],
    lead_term: float,
    upsampling: int = -1,
) -> torch.Tensor:
    """Simulates the scattering potentials for each atom around its neighborhood.

    Parameters
    ----------
    atom_positions_zyx : torch.Tensor
        The atom coordinates in Angstroms. Shape (N, 3) where N is the number
        of atoms.
    atom_ids : list[str]
        The atom IDs as a list of uppercase element symbols.
    atom_b_factors : torch.Tensor
        The atom B factors.
    sim_pixel_spacing : float
        The pixel spacing for the final simulation in Angstroms.
    sim_volume_shape : tuple[int, int, int]
        The shape of the final simulation volume.
    lead_term : float
        The lead term for the scattering potential.
    upsampling : int
        The upsampling factor. If -1, the upsampling factor is calculated
        automatically.


    """
    actual_upsampling, upsampled_pixel_size, upsampled_shape = _setup_sim3d_upsampling(
        sim_pixel_spacing=sim_pixel_spacing,
        sim_volume_shape=sim_volume_shape,
        upsampling=upsampling,
    )

    mean_b_factor = torch.mean(atom_b_factors)
    atom_indices, atom_dds, voxel_offsets_flat = _setup_upsampling_coords(
        atom_positions_zyx=atom_positions_zyx,
        upsampled_pixel_size=upsampled_pixel_size,
        upsampled_shape=upsampled_shape,
        mean_b_factor=mean_b_factor,
    )

    # Reshaping tensors for future broadcasting
    atom_pos = einops.rearrange(atom_indices, "a d -> a 1 d")
    atom_dds = einops.rearrange(atom_dds, "a d -> a 1 d")
    voxel_offsets_flat = einops.rearrange(voxel_offsets_flat, "v d -> 1 v d")

    # Calculate voxel positions relative to atom center
    voxel_positions = atom_pos + voxel_offsets_flat  # (n_atoms, h*w*d, 3)

    # Calculate relative coordinates of each voxel in the neighborhood
    relative_coords = (
        voxel_positions - atom_pos - atom_dds - PIXEL_OFFSET
    ) * upsampled_pixel_size
    coords1 = relative_coords
    coords2 = relative_coords + upsampled_pixel_size

    ########################################
    ### Scattering potential calculation ###
    ########################################

    neighborhood_potentials = get_scattering_potential_of_voxel_batch(
        zyx_coords1=coords1,  # shape(a, v, d)
        zyx_coords2=coords2,  # shape(a, v, d)
        atom_ids=atom_ids,
        atom_b_factors=atom_b_factors,
        lead_term=lead_term,
    )

    return {
        "neighborhood_potentials": neighborhood_potentials,
        "voxel_positions": voxel_positions,
        "upsampled_shape": upsampled_shape,
        "upsampled_pixel_size": upsampled_pixel_size,
        "actual_upsampling": actual_upsampling,
    }


def place_voxel_neighborhoods_in_volume(
    neighborhood_potentials: torch.Tensor,  # shape (N, h*w*d)
    voxel_positions: torch.LongTensor,  # shape (N, h*w*d, 3)
    final_volume: torch.Tensor,  # shape (H, W, D)
) -> torch.Tensor:
    """Places pre-calculate voxels of (N, h, w, d) into the volume (H, W, D).

    Parameters
    ----------
    neighborhood_potentials : torch.Tensor
        The pre-calculated scattering potentials in voxel neighborhoods around
        each atom. Shape (N, h*w*d).
    voxel_positions : torch.LongTensor
        The voxel offset positions for each of the neighborhoods. Shape (N, 3)
        with last dim representing (x, y z).
    final_volume : torch.Tensor
        The final volume to place the neighborhoods into. Shape (H, W, D).

    Returns
    -------
    torch.Tensor
        The final volume with the neighborhoods placed in.
    """
    index_positions = voxel_positions.long()
    final_volume.index_put_(
        indices=(
            index_positions[:, :, 0],
            index_positions[:, :, 1],
            index_positions[:, :, 2],
        ),
        values=neighborhood_potentials,
        accumulate=True,
    )

    return final_volume


def calculate_simulation_dose_filter_3d(
    shape: tuple[int, int, int],
    dose_start: float,
    dose_end: float,
    critical_bfactor: float,
    modify_signal: Literal["None", "sqrt", "rel_diff"],
    rfft: bool = True,
    fftshift: bool = False,
) -> torch.Tensor:
    """Helper function to calculate a cumulative dose filter for a simulation.

    Parameters
    ----------
    shape : tuple[int, int, int]
        The requested return shape of the dose filter.
    dose_start : float
        The starting dose for the dose filter in units of e-/A^2.
    dose_end : float
        The ending dose for the dose filter in units of e-/A^2.
    critical_bfactor : float
        The critical B factor for the dose filter. If -1, the Grant and
        Grigorieff (2015) critical exposure dose weighting is used.
    modify_signal : Literal["None", "sqrt", "rel_diff"]
        The method to modify the signal after applying the dose filter. Options
        are "None", "sqrt", and "rel_diff".
        - 'None': No modification is applied.'
        - 'sqrt': The square root of the filter is applied.
        - 'rel_diff': Filter becomes 1 - (1 - filter) / (1 + filter).
    rfft : bool
        If True, the filter is returned in rfft format. Default is True.
    fftshift : bool
        If True, the filter is fftshifted. Default is False.

    Returns
    -------
    torch.Tensor
    """
    dose_filter = cumulative_dose_filter_3d(
        volume_shape=shape,
        start_exposure=dose_start,
        end_exposure=dose_end,
        crit_exposure_bfactor=critical_bfactor,
        rfft=rfft,
        fftshift=fftshift,
    )

    if modify_signal == "None":
        pass
    elif modify_signal == "sqrt":
        dose_filter = torch.sqrt(dose_filter)
    elif modify_signal == "rel_diff":
        denominator = 1 + dose_filter
        epsilon = 1e-10
        denominator = torch.clamp(denominator, min=epsilon)
        modification = 1 - (1 - dose_filter) / denominator
        dose_filter = modification

    return dose_filter


def apply_simulation_filters(
    upsampled_volume: torch.Tensor,
    actual_upsampling: int,
    final_shape: tuple[int, int, int],
    apply_dose_weighting: bool,
    dose_start: float,
    dose_end: float,
    dose_filter_modify_signal: Literal["None", "sqrt", "rel_diff"],
    dose_filter_critical_bfactor: float,
    apply_dqe: bool,
    mtf_frequencies: torch.Tensor,
    mtf_amplitudes: torch.Tensor,
) -> torch.Tensor:
    """Apply filters to the simulated volume.

    This function does the following:
    1. Apply dose weighting in the upsampled Fourier volume, if requested
    2. Fourier crop back to the desired size (final_shape)
    3. Apply DQE to the volume in Fourier space
    4. Inverse FFT to get the final real-space simulated volume

    Parameters
    ----------
    upsampled_volume : torch.Tensor
        The upsampled volume in Fourier space.
    actual_upsampling : int
        The actual upsampling factor used for the simulation.
    final_shape : tuple[int, int, int]
        The final shape of the simulated volume.
    apply_dose_weighting : bool
        If True, apply dose weighting to the simulation.
    dose_start : float
        The starting dose for the dose weighting filter in units of e-/A^2.
    dose_end : float
        The ending dose for the dose weighting filter in units of e-/A^2.
    dose_filter_modify_signal : Literal["None", "sqrt", "rel_diff"]
        The method to modify the signal after applying the dose filter. Options
        are "None", "sqrt", and "rel_diff".
    dose_filter_critical_bfactor : float
        The critical B factor for the dose filter. If -1, the Grant and
        Grigorieff (2015) critical exposure dose weighting is used.
    apply_dqe : bool
        If True, apply the detective quantum efficiency (DQE) filter to the
        simulation. Applied in regular (not upsampled) Fourier space.
    mtf_frequencies : torch.Tensor
        The frequencies for the modulation transfer function (MTF) filter in
        units of inverse pixels ranging from 0 to 0.5.
    mtf_amplitudes : torch.Tensor
        The amplitudes for the modulation transfer function (MTF) filter at the
        corresponding frequencies.

    Returns
    -------
    torch.Tensor
        The final simulated volume.

    """
    upsampled_volume_rfft = torch.fft.rfftn(upsampled_volume, dim=(-3, -2, -1))
    upsampled_shape = upsampled_volume.shape

    # Calculate and apply dose filter, if requested
    if apply_dose_weighting:
        dose_filter = calculate_simulation_dose_filter_3d(
            shape=upsampled_shape,
            dose_start=dose_start,
            dose_end=dose_end,
            critical_bfactor=dose_filter_critical_bfactor,
            modify_signal=dose_filter_modify_signal,
            rfft=True,
            fftshift=False,
        )
        upsampled_volume_rfft *= dose_filter

    # Fourier crop back to desired size
    if actual_upsampling != 1:
        upsampled_volume_rfft = fourier_rescale_3d_force_size(
            volume_fft=upsampled_volume_rfft,
            volume_shape=upsampled_shape,
            target_size=final_shape[0],  # TODO: pass this arg as a tuple
            rfft=True,
            fftshift=False,
        )

    # Apply DQE, if requested
    if apply_dqe:
        mtf = make_mtf_grid(
            image_shape=final_shape,
            mtf_frequencies=mtf_frequencies,
            mtf_amplitudes=mtf_amplitudes,
            rfft=True,
            fftshift=False,
        )
        upsampled_volume_rfft *= mtf

    # Inverse FFT
    cropped_volume = torch.fft.irfftn(
        upsampled_volume_rfft,
        s=final_shape,
        dim=(-3, -2, -1),
    )
    # NOTE: ifftshift not needed since volume here was never fftshifted
    # cropped_volume = torch.fft.ifftshift(cropped_volume, dim=(-3, -2, -1))

    return cropped_volume


def simulate3d(
    atom_positions_zyx: torch.Tensor,
    atom_ids: list[str],
    atom_b_factors: torch.Tensor,
    beam_energy_kev: float,
    sim_pixel_spacing: float,
    sim_volume_shape: tuple[int, int, int],
    requested_upsampling: int = -1,
    apply_dose_weighting: bool = True,
    dose_start: float = 0.0,
    dose_end: float = 30.0,
    dose_filter_modify_signal: Literal["None", "sqrt", "rel_diff"] = "None",
    dose_filter_critical_bfactor: float = -1,
    apply_dqe: bool = False,
    mtf_frequencies: torch.Tensor = None,
    mtf_amplitudes: torch.Tensor = None,
    # gpu_ids: int | list[int] = -999,  # TODO: implement gpu selection
) -> torch.Tensor:
    """Simulate 3D electron scattering volume with requested parameters.

    Parameters
    ----------
    atom_positions_zyx : torch.Tensor
        The atom positions in Angstroms. Shape (N, 3) where N is the number of
        atoms.
    atom_ids : list[str]
        The atomic IDs of each atom in the model.
    atom_b_factors : torch.Tensor
        The B factors of each atom in the model.
    beam_energy_kev : float
        The electron beam energy in kV.
    sim_pixel_spacing : float
        The pixel spacing for the final simulation in Angstroms.
    sim_volume_shape : tuple[int, int, int]
        The shape of the final simulation volume. Center of volume is at
        the origin (0.0, 0.0, 0.0).
    requested_upsampling : int
        The upsampling factor to use for the simulation. If -1, the upsampling
        factor is calculated automatically.
    apply_dose_weighting : bool
        If True, apply dose weighting to the simulation.
    dose_start : float
        The starting dose for the dose weighting filter in units of e-/A^2.
        Default is 0.0.
    dose_end : float
        The ending dose for the dose weighting filter in units of e-/A^2.
        Default is 30.0.
    dose_filter_modify_signal : str
        The method to modify the signal after applying the dose filter. Options
        are "None", "sqrt", and "rel_diff". Default is "None".
        - 'None': No modification is applied.
        - 'sqrt': The square root of the filter is applied.
        - 'rel_diff': Filter becomes 1 - (1 - filter) / (1 + filter).
    dose_filter_critical_bfactor : float
        The critical B factor for the dose filter. Default is -1, which uses
        the Grant and Grigorieff (2015) critical exposure dose weighting.
    apply_dqe : bool
        If True, apply the detective quantum efficiency (DQE) filter to the
        simulation. If True, both 'mtf_frequencies' and 'mtf_amplitudes' must
        be provided. Default is False.
    mtf_frequencies : torch.Tensor
        The frequencies for the modulation transfer function (MTF) filter in
        units of inverse pixels ranging from 0 to 0.5. Required if 'apply_dqe'
        is True.
    mtf_amplitudes : torch.Tensor
        The amplitudes for the modulation transfer function (MTF) filter at the
        corresponding frequencies. Must be the same length as
        'mtf_frequencies'. Required if 'apply_dqe' is True.
    # gpu_ids : int | list[int]

    Returns
    -------
    torch.Tensor
        The simulated 3D volume in real space.
    """
    # Validate portions of the input before continuing
    _validate_dose_filter_inputs(
        dose_filter_modify_signal, dose_filter_critical_bfactor
    )
    _validate_dqe_filter_inputs(apply_dqe, mtf_frequencies, mtf_amplitudes)

    # Select devices for calculation
    # TODO: Implement GPU selection

    # Calculate the atom-wise scattering potentials
    lead_term = _calculate_lead_term(beam_energy_kev, sim_pixel_spacing)
    scattering_results = simulate_atomwise_scattering_potentials(
        atom_positions_zyx=atom_positions_zyx,
        atom_ids=atom_ids,
        atom_b_factors=atom_b_factors,
        sim_pixel_spacing=sim_pixel_spacing,
        sim_volume_shape=sim_volume_shape,
        lead_term=lead_term,
        upsampling=requested_upsampling,
    )

    neighborhood_potentials = scattering_results["neighborhood_potentials"]
    voxel_positions = scattering_results["voxel_positions"]
    upsampled_shape = scattering_results["upsampled_shape"]
    actual_upsampling = scattering_results["actual_upsampling"]

    # Place the voxel neighborhoods in the final, upsampled volume
    upsampled_volume = torch.zeros(upsampled_shape, dtype=torch.float32)
    upsampled_volume = place_voxel_neighborhoods_in_volume(
        neighborhood_potentials=neighborhood_potentials,
        voxel_positions=voxel_positions,
        final_volume=upsampled_volume,
    )

    # Apply filters to the simulation
    final_volume = apply_simulation_filters(
        upsampled_volume=upsampled_volume,
        actual_upsampling=actual_upsampling,
        final_shape=sim_volume_shape,
        apply_dose_weighting=apply_dose_weighting,
        dose_start=dose_start,
        dose_end=dose_end,
        dose_filter_modify_signal=dose_filter_modify_signal,
        dose_filter_critical_bfactor=dose_filter_critical_bfactor,
        apply_dqe=apply_dqe,
        mtf_frequencies=mtf_frequencies,
        mtf_amplitudes=mtf_amplitudes,
    )

    return final_volume



"""Test for functions defined in src/ttsim3d/simulate3d.py"""

import pytest
import torch

from ttsim3d.simulate3d import _validate_dose_filter_inputs, _validate_dqe_filter_inputs


##################################
### Tests for helper functions ###
##################################
def test_calculate_lead_term():
    pass


def test_validate_dose_filter_inputs():
    """Basic tests for validate_dose_filter_inputs method."""
    # Ensure no errors are raised when inputs are valid
    permitted_modify_signals = ["None", "sqrt", "rel_diff"]
    for mf in permitted_modify_signals:
        _validate_dose_filter_inputs(
            dose_filter_modify_signal=mf,
            dose_filter_critical_bfactor=-1,  # Grant-Grigorieff 2015
        )
        _validate_dose_filter_inputs(
            dose_filter_modify_signal=mf,
            dose_filter_critical_bfactor=30.0,
        )

    # Ensure errors are raised when inputs are invalid
    with pytest.raises(ValueError):
        _validate_dose_filter_inputs(
            dose_filter_modify_signal="invalid",  # bad parameter
            dose_filter_critical_bfactor=-1,
        )

    with pytest.raises(ValueError):
        _validate_dose_filter_inputs(
            dose_filter_modify_signal="None",
            dose_filter_critical_bfactor=-2.5,  # bad parameter
        )


def test_validate_dqe_filter_inputs():
    """Basic tests for validate_dqe_filter_inputs method."""
    _validate_dqe_filter_inputs(
        apply_dqe=False,
        mtf_frequencies=None,
        mtf_amplitudes=None,
    )
    _validate_dqe_filter_inputs(
        apply_dqe=True,
        mtf_frequencies=torch.tensor([0.0, 0.25, 0.5]),
        mtf_amplitudes=torch.tensor([1.0, 0.5, 0.0]),
    )

    # Errors raised when mtf tensors are none
    with pytest.raises(ValueError):
        _validate_dqe_filter_inputs(
            apply_dqe=True,
            mtf_frequencies=None,
            mtf_amplitudes=torch.tensor([1.0, 0.5, 0.0]),
        )
    with pytest.raises(ValueError):
        _validate_dqe_filter_inputs(
            apply_dqe=True,
            mtf_frequencies=torch.tensor([0.0, 0.25, 0.5]),
            mtf_amplitudes=None,
        )

    # Errors raised when mtf tensors are not the same size
    with pytest.raises(ValueError):
        _validate_dqe_filter_inputs(
            apply_dqe=True,
            mtf_frequencies=torch.tensor([0.0, 0.25, 0.5]),
            mtf_amplitudes=torch.tensor([1.0, 0.5]),
        )


def test_setup_sim3d_upsampling():
    pass


def test_setup_upsampling_coords():
    pass


###############################################
### Tests for the core simulation functions ###
###############################################
def test_simulate_atomwise_scattering_potentials():
    pass


def test_place_voxel_neighborhoods_in_volume():
    pass


def test_calculate_simulation_dose_filter_3d():
    pass


def test_apply_simulation_filters():
    pass


def test_simulate3d():
    pass



"""Pydantic models for input parameters."""

import os
import pathlib
from typing import Annotated, Any, Literal, Optional

import torch
from pydantic import BaseModel, ConfigDict, Field, field_validator
from pydantic.json_schema import SkipJsonSchema
from torch_fourier_filter.mtf import read_mtf

from ttsim3d.mrc_handler import tensor_to_mrc
from ttsim3d.pdb_handler import load_model, remove_hydrogens
from ttsim3d.simulate3d import ALLOWED_DOSE_FILTER_MODIFICATIONS, simulate3d

DEFAULT_MTF_REFERENCES = {
    "de20_300kv": "src/data/mtf_de20_300kV.star",
    "falcon2_300kv": "src/data/mtf_falcon2_300kV.star",
    "falcon3EC_200kv": "src/data/mtf_falcon3EC_200kV.star",
    "falcon3EC_300kv": "src/data/mtf_falcon3EC_300kV.star",
    "falcon4EC_200kv": "src/data/mtf_falcon4EC_200kV.star",
    "falcon4EC_300kv": "src/data/mtf_k2_300kV_FL2.star",
    "k2_300kv": "src/data/mtf_k2_300kV.star",
    "k2_200kV_FL2": "src/data/mtf_k2_200kV_FL2.star",
    "k2_300kV_FL2": "src/data/mtf_k2_300kV_FL2.star",
    "k3_200kV_FL2": "src/data/mtf_standard_k3_200kV_FL2.star",
    "k3_300kV_FL2": "src/data/mtf_standard_k3_300kV_FL2.star",
}

# Pydantic type annotation for large tensor excluded from JSON schema and dump
ExcludedTensor = SkipJsonSchema[
    Annotated[torch.Tensor, Field(default=None, exclude=True)]
]


class SimulatorConfig(BaseModel):
    """Configuration for simulating a 3D volume.

    These simulation parameters are intended to be model agnostic, that is,
    the configuration can be used between multiple structure simulations.

    Most model parameters are related to simulation filters to apply with a
    handful also included for storing intermediate results during the
    calculation.

    Attributes
    ----------
    voltage : float
        The voltage of the microscope in kV. Default is 300 kV.
    apply_dose_weighting : bool
        If True, apply dose weighting to the simulation.
    crit_exposure_bfactor : float
        B-factor to use in critical exposure calculations. The default is -1
        and corresponds to the fitted critical exposure function in Grant and
        Grigorieff, 2015.
    dose_filter_modify_signal : Literal["None", "sqrt", "rel_diff"]
        Signal modification to apply to the dose filter. Currently supports
        - 'None': No modification
        - 'sqrt': x' = sqrt(x)
        - 'rel_diff': x' = 1 - (1 - x) / (1 - x)
    dose_start : float
        The starting dose in e/A^2. The default is 0.0 e/A^2.
    dose_end : float
        The ending dose in e/A^2. The default is 30.0 e/A^2.
    apply_dqe : bool
        If True, apply a DQE filter to the simulation.
    mtf_reference : str
        Path to the modulation transfer function (MTF) reference star file, or
        one of the known MTF reference files in:
        - 'k2_300kV': The MTF reference for a K2 camera at 300 kV.
    upsampling : int
        The upsampling factor to apply to the simulation. The default is -1 and
        corresponds to automatic calculation of the upsampling factor.
    store_volume : bool
        If True, store the final simulated volume in real space after requested
        simulation filters are applied under the attribute `Simulator.volume`.
        Default is True.

    Methods
    -------
    model_dump -> dict
    """

    # Serializable attributes
    voltage: Annotated[float, Field(ge=0.0)] = 300.0
    apply_dose_weighting: Annotated[bool, Field(default=True)] = True
    crit_exposure_bfactor: Annotated[float, Field(default=-1)] = -1
    dose_filter_modify_signal: Annotated[
        Literal["None", "sqrt", "rel_diff"], Field(default="None")
    ] = "None"
    dose_start: Annotated[float, Field(ge=0.0)] = 0.0
    dose_end: Annotated[float, Field(ge=0.0)] = 30.0
    apply_dqe: Annotated[bool, Field(default=True)] = True
    mtf_reference: Annotated[str, Field(default="k2_300kV")] = "k2_300kV"
    upsampling: Annotated[int, Field(default=-1)] = -1

    store_volume: Annotated[bool, Field(default=True)] = True

    @field_validator("crit_exposure_bfactor")  # type: ignore
    def validate_crit_exposure_bfactor(cls, v):
        """Validate model input `crit_exposure_bfactor`."""
        if not isinstance(v, (float, int)):
            e = f"Invalid critical exposure B-factor of type: {v}."
            e += "Expected a float or integer."
            raise ValueError(e)

        return v

    @field_validator("dose_filter_modify_signal")  # type: ignore
    def validate_dose_filter_modify_signal(cls, v):
        """Validate model input `dose_filter_modify_signal`."""
        if v not in ALLOWED_DOSE_FILTER_MODIFICATIONS:
            e = f"Invalid dose filter signal modification: {v}."
            e += f"Allowed values are: {ALLOWED_DOSE_FILTER_MODIFICATIONS}"
            raise ValueError(e)
        return v

    @field_validator("mtf_reference")  # type: ignore
    def validate_mtf_reference(cls, v):
        """Validate model input `mtf_reference`."""
        _path_exists = os.path.exists(v)
        _is_default = v in DEFAULT_MTF_REFERENCES

        if not _path_exists and not _is_default:
            e = f"Invalid MTF reference file: {v}. "
            e += "Please provide a valid path to an MTF reference file."
            raise ValueError(e)

        if _is_default:
            return DEFAULT_MTF_REFERENCES[v]

        return v

    @property
    def mtf_tensors(self) -> tuple[torch.Tensor, torch.Tensor]:
        """Returns the MTF tensors from the reference file."""
        frequencies, amplitudes = read_mtf(file_path=self.mtf_reference)

        return frequencies, amplitudes

    # def model_dump(self) -> dict:
    #     """Return the model as a dictionary."""
    #     return self.dict()


class Simulator(BaseModel):
    """Class for simulating a 3D volume from a atomistic structure.

    Attributes
    ----------
    pixel_spacing : float
        The pixel spacing of the simulated volume in units of Angstroms. Must
        be greater than 0, and defaults to 1.0 Angstroms.
    volume_shape : tuple[int, int, int]
        The shape of the simulated volume in pixels. The default is
        (400, 400, 400).
    pdb_filepath : pathlib.Path
        The path to the PDB file containing the atomic structure to simulate.
    b_factor_scaling : float
        The scaling factor to apply to the B-factors of the atoms in the pdb
        file. The default is 1.0.
    additional_b_factor : float
        Additional B-factor to apply to the atoms in the pdb file. The default
        is 0.0.
    simulator_config : SimulatorConfig
        Simulation configuration.
    atom_positions_zyx : torch.Tensor
        The positions (float tensor) of the atoms in the structure in units of
        Angstroms. Non-serializable attribute.
    atom_identities : torch.Tensor
        The atomic identities (str tensor) of the atoms in the structure.
        Non-serializable attribute.
    atom_b_factors : torch.Tensor
        The B-factors (float tensor) of the atoms in the structure in units of
        A^2. Non-serializable attribute.
    volume : torch.Tensor
        The simulated volume in real space after requested simulation filters
        are applied. Non-serializable attribute. Only stored if requested in
        the SimulatorConfig.

    Methods
    -------
    __init__ -> None
    load_atoms_from_pdb_model -> None
        Loads the structure atoms from the held pdb file into the attributes
        `atom_positions_zyx`, `atom_identities`, and `atom_b_factors`.
    get_scale_atom_b_factors -> torch.Tensor
        Returns the scaled b-factors of the atoms in the structure based on
        the attributes `b_factor_scaling` and `additional_b_factor`.
    run -> torch.Tensor
        Runs the simulation and returns the simulated volume.
    export_to_mrc -> None
        Exports the simulated volume to an MRC file.
    """

    # Pydantic model configuration
    model_config = ConfigDict(arbitrary_types_allowed=True)

    # Serializable attributes
    pixel_spacing: Annotated[float, Field(ge=0.0)] = 1.0
    volume_shape: Annotated[tuple[int, int, int], Field(default=(400, 400, 400))] = (
        400,
        400,
        400,
    )
    pdb_filepath: Annotated[pathlib.Path, Field(...)]
    b_factor_scaling: Annotated[float, Field(default=1.0)] = 1.0
    additional_b_factor: Annotated[float, Field(default=0.0)] = 0.0
    simulator_config: SimulatorConfig

    # Non-serializable and schema-excluded attributes
    atom_positions_zyx: ExcludedTensor
    atom_identities: ExcludedTensor
    atom_b_factors: ExcludedTensor
    volume: ExcludedTensor

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

        self.load_atoms_from_pdb_model()

    def load_atoms_from_pdb_model(self) -> None:
        """Loads the structure atoms from held pdb file."""
        atom_positions_zyx, atom_ids, atom_b_factors = load_model(self.pdb_filepath)
        atom_positions_zyx, atom_ids, atom_b_factors = remove_hydrogens(
            atom_positions_zyx, atom_ids, atom_b_factors
        )

        self.atom_positions_zyx = atom_positions_zyx
        self.atom_identities = atom_ids
        self.atom_b_factors = atom_b_factors

    def get_scale_atom_b_factors(self) -> torch.Tensor:
        """Returns b-factors transformed by the scale and additional b-factor.

        Parameters
        ----------
        None

        Returns
        -------
        b_fac: torch.Tensor
            The scaled b-factors.
        """
        if self.atom_b_factors is None:
            raise ValueError("No atom B-factors loaded.")

        b_fac = self.atom_b_factors * self.b_factor_scaling
        b_fac += self.additional_b_factor

        # NOTE (Josh): cisTEM includes a 0.25 factor. Unsure why, but keeping
        # consistent with the original implementation for now.
        b_fac *= 0.25

        return b_fac

    def run(
        self,
        gpu_ids: Optional[int | list[int]] = None,
        atom_indices: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Runs the simulation and returns the simulated volume.

        Parameters
        ----------
        gpu_ids: int | list[int]
            A list of GPU IDs to use for the simulation. The default is 'None'
            which will use the CPU. A value of '-1' will use all available
            GPUs, otherwise a list of integers greater than or equal to 0 are
            expected.
        atom_indices: torch.Tensor
            The indices of the atoms to simulate. The default is 'None' which
            will simulate all atoms in the structure.

        Returns
        -------
        volume: torch.Tensor
            The simulated volume.
        """
        assert self.atom_positions_zyx is not None, "No atom positions loaded."
        assert self.atom_identities is not None, "No atom identities loaded."

        if atom_indices is None:
            atom_indices = torch.arange(self.atom_identities.size(0))

        # Select GPUs to use, or use CPU
        # TODO: Implement GPU selection

        # Get the scaled atom b-factors
        atom_b_factors = self.get_scale_atom_b_factors()

        # Calculate the mtf_frequencies and mtf_amplitudes from reference file
        mtf_frequencies, mtf_amplitudes = self.simulator_config.mtf_tensors

        volume = simulate3d(
            atom_positions_zyx=self.atom_positions_zyx,
            atom_ids=self.atom_identities,
            atom_b_factors=atom_b_factors,
            beam_energy_kev=self.simulator_config.voltage,
            sim_pixel_spacing=self.pixel_spacing,
            sim_volume_shape=self.volume_shape,
            requested_upsampling=self.simulator_config.upsampling,
            apply_dose_weighting=self.simulator_config.apply_dose_weighting,
            dose_start=self.simulator_config.dose_start,
            dose_end=self.simulator_config.dose_end,
            dose_filter_modify_signal=self.simulator_config.dose_filter_modify_signal,
            dose_filter_critical_bfactor=self.simulator_config.crit_exposure_bfactor,
            apply_dqe=self.simulator_config.apply_dqe,
            mtf_frequencies=mtf_frequencies,
            mtf_amplitudes=mtf_amplitudes,
        )

        if self.simulator_config.store_volume:
            self.volume = volume

        return volume

    def export_to_mrc(
        self,
        mrc_filepath: str | os.PathLike,
        gpu_ids: Optional[int | list[int]] = None,
        atom_indices: Optional[torch.Tensor] = None,
    ) -> None:
        """Exports the simulated volume to an MRC file.

        Parameters
        ----------
        mrc_filepath: str | os.PathLike
            The file path to save the MRC file.
        gpu_ids: int | list[int]
            A list of GPU IDs to use for the simulation. The default is 'None'
            which will use the CPU. A value of '-1' will use all available
            GPUs, otherwise a list of integers greater than or equal to 0 are
            expected. The default is 'None'. This is passed to the `run`
            method.
        atom_indices: torch.Tensor
            The indices of the atoms to simulate. The default is 'None' which
            will simulate all atoms in the structure. This is passed to the
            `run` method.

        Returns
        -------
        None
        """
        volume = self.run(gpu_ids=gpu_ids, atom_indices=atom_indices)

        tensor_to_mrc(
            output_filename=str(mrc_filepath),
            final_volume=volume,
            sim_pixel_spacing=self.pixel_spacing,
        )



"""Deals with grid coordinates."""

import einops
import torch


def get_upsampling(
    wanted_pixel_size: float, wanted_output_size: int, max_size: int = 1536
) -> int:
    """Calculate the upsampling factor for the simulation volume.

    Parameters
    ----------
    wanted_pixel_size : float
        The pixel size in Angstroms.
    wanted_output_size : float
        The output size of the cubic volume.
    max_size : int
        Optional maximum size of the volume. Default is 1536.

    Returns
    -------
    int
        The upsampling factor.
    """
    if wanted_pixel_size > 1.5 and wanted_output_size * 4 < max_size:
        print("Oversampling your 3d by a factor of 4 for calculation.")
        return 4

    if 0.75 < wanted_pixel_size <= 1.5 and wanted_output_size * 2 < max_size:
        print("Oversampling your 3d by a factor of 2 for calculation.")
        return 2

    return 1


def get_atom_voxel_indices(
    atom_zyx: torch.Tensor,
    upsampled_pixel_size: float,
    upsampled_shape: tuple[int, int, int],
    offset: float = 0.5,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Calculate the voxel indices of the atoms.

    Parameters
    ----------
    atom_zyx : torch.Tensor
        The atom coordinates in Angstroms.
    upsampled_pixel_size : float
        The pixel size in Angstroms.
    upsampled_shape : tuple[int, int, int]
        The shape of the upsampled volume.
    offset : float
        Optional voxel edge offset in units of voxels. Default is 0.5.

    Returns
    -------
    tuple[torch.Tensor,torch.Tensor]
        The voxel indices and the offset from the edge of the voxel.
    """
    origin_idx = (
        upsampled_shape[0] / 2,
        upsampled_shape[1] / 2,
        upsampled_shape[2] / 2,
    )
    this_coords = (
        (atom_zyx / upsampled_pixel_size)
        + torch.tensor(origin_idx).unsqueeze(0)
        + offset
    )
    atom_indices = torch.floor(this_coords)  # these are the voxel indices
    atom_dds = (
        this_coords - atom_indices - offset
    )  # this is offset from the edge of the voxel

    return atom_indices, atom_dds


def get_size_neighborhood_cistem(
    mean_b_factor: float, upsampled_pixel_size: float
) -> int:
    """Calculate the size of the neighborhood of voxels (mirrors cisTEM).

    Parameters
    ----------
    mean_b_factor : float
        The mean B factor over all the atoms.
    upsampled_pixel_size : float
        The pixel size in Angstroms.

    Returns
    -------
    int
        The size (number of voxels in one direction) of the neighborhood.
    """
    tmp = 0.4 * (0.6 * mean_b_factor) ** 0.5 + 0.2
    tmp = torch.round(tmp / upsampled_pixel_size)

    return int(tmp + 1)


def get_voxel_neighborhood_offsets(
    mean_b_factor: float, upsampled_pixel_size: float
) -> torch.Tensor:
    """Offset arrays for the voxel neighborhood.

    Calculate the offsets of the voxel neighborhood. Returned as a flat tensor
    with shape (n^3, 3) where n is the size of the neighborhood in one dimension.

    Parameters
    ----------
    mean_b_factor : float
        The mean B factor of the atoms.
    upsampled_pixel_size : float
        The pixel size in Angstroms.

    Returns
    -------
    torch.Tensor
        The offsets of the voxel neighborhood.

    """
    # Get the size of the voxel neighbourhood to calculate the potential of each atom
    size_neighborhood = get_size_neighborhood_cistem(
        mean_b_factor, upsampled_pixel_size
    )
    neighborhood_range = torch.arange(-size_neighborhood, size_neighborhood + 1)
    # Create coordinate grids for the neighborhood
    sz, sy, sx = torch.meshgrid(
        neighborhood_range, neighborhood_range, neighborhood_range, indexing="ij"
    )
    voxel_offsets = torch.stack([sz, sy, sx])  # (3, n, n, n)
    # Flatten while preserving the relative positions
    voxel_offsets_flat = einops.rearrange(
        voxel_offsets, "c x y z -> (x y z) c"
    )  # (n^3, 3)

    return voxel_offsets_flat


# This will definitely be moved to a different program
def fourier_rescale_3d_force_size(
    volume_fft: torch.Tensor,
    volume_shape: tuple[int, int, int],
    target_size: int,
    rfft: bool = True,
    fftshift: bool = False,
) -> torch.Tensor:
    """
    Crop a 3D Fourier-transformed volume to a specific target size.

    Parameters
    ----------
    volume_fft : torch.Tensor
        The Fourier-transformed volume.
    volume_shape : tuple[int, int, int]
        The original shape of the volume.
    target_size : int
        The target size of the cropped volume.
    rfft : bool
        Whether the input is a real-to-complex Fourier Transform.
    fftshift : bool
        Whether the zero frequency is shifted to the center.

    Returns
    -------
    cropped_fft_shifted_back : torch.Tensor
        The cropped fft
    """
    # Ensure the target size is even
    assert target_size > 0, "Target size must be positive."

    # Get the original size of the volume
    assert (
        volume_shape[0] == volume_shape[1] == volume_shape[2]
    ), "Volume must be cubic."

    # Step 1: Perform real-to-complex Fourier Transform (rfftn)
    # and shift the zero frequency to the center
    if not fftshift:
        volume_fft = torch.fft.fftshift(
            volume_fft, dim=(-3, -2, -1)
        )  # Shift along first two dimensions only

    # Calculate the dimensions of the rfftn output
    rfft_size_z, rfft_size_y, rfft_size_x = volume_fft.shape

    # Calculate cropping indices for each dimension
    center_z = rfft_size_z // 2
    center_y = rfft_size_y // 2
    center_x = rfft_size_x // 2

    # Define the cropping ranges
    crop_start_z = int(center_z - target_size // 2)
    crop_end_z = int(crop_start_z + target_size)
    crop_start_y = int(center_y - target_size // 2)
    crop_end_y = int(crop_start_y + target_size)
    crop_start_x = int(center_x - target_size // 2)
    crop_end_x = int(
        target_size // 2 + 1
    )  # Crop from the high-frequency end only along the last dimension

    # Step 2: Crop the Fourier-transformed volume
    cropped_fft = torch.zeros_like(volume_fft)
    if rfft:
        cropped_fft = volume_fft[
            crop_start_z:crop_end_z, crop_start_y:crop_end_y, -crop_end_x:
        ]
    else:
        crop_end_x = int(crop_start_x + target_size)
        cropped_fft = volume_fft[
            crop_start_z:crop_end_z, crop_start_y:crop_end_y, crop_start_x:crop_end_x
        ]

    # Step 3: Inverse shift and apply the inverse rFFT to return to real space
    cropped_fft_shifted_back = torch.fft.ifftshift(cropped_fft, dim=(-3, -2))

    return cropped_fft_shifted_back



"""Comparison of simulation to static simulated .mrc file.

NOTE: If the core algorithm for the simulation changes, then this test will
fail and the reference "good" simulated structure will need to be updated.
"""

import os
from pathlib import Path

import mrcfile
import numpy as np
import pytest
import requests
from torch_fourier_filter.mtf import read_mtf

from ttsim3d.pdb_handler import load_model, remove_hydrogens
from ttsim3d.simulate3d import simulate3d

# Remote filepaths for testing
PDB_STRUCTURE_FILEPATH = "https://zenodo.org/records/14219436/files/parsed_6Q8Y_whole_LSU_match3.pdb?download=1"
SIMULATED_MRC_FILEPATH = "https://zenodo.org/records/14219436/files/parsed_6Q8Y_whole_LSU_match3_37dff6e22103f08d3453b0163435476c2808d08b.mrc?download=1"
DQE_STARFILE_FILEPATH = (
    "https://zenodo.org/records/14219436/files/mtf_k2_300kV.star?download=1"
)


# Temporary paths
this_path = Path(__file__).resolve()
this_dir = this_path.parent
tmp_dir = this_dir / "tmp"


def download_file(url: str, dest_folder: str) -> str:
    """Download a file from a URL to a destination folder."""
    if not os.path.exists(dest_folder):
        os.makedirs(dest_folder)

    filename = url.split("/")[-1].split("?")[0]
    file_path = os.path.join(dest_folder, filename)

    # Skip download if file already exists
    if os.path.exists(file_path):
        return file_path

    response = requests.get(url, stream=True)
    response.raise_for_status()  # Raise an error for bad status codes
    # total_size = int(response.headers.get("content-length", 0))
    block_size = 1024  # 1 Kibibyte

    with open(file_path, "wb") as file:
        for data in response.iter_content(block_size):
            file.write(data)

    # print(f"Downloaded {file_path}")
    return file_path


def setup_simulation():
    """Download necessary files and setup temporary directory."""
    # print(this_path)
    # print(this_dir)
    # print(tmp_dir)

    pdb_filepath = download_file(PDB_STRUCTURE_FILEPATH, tmp_dir)
    mrc_filepath = download_file(SIMULATED_MRC_FILEPATH, tmp_dir)
    dqe_filepath = download_file(DQE_STARFILE_FILEPATH, tmp_dir)

    return pdb_filepath, mrc_filepath, dqe_filepath


def get_simulation_kwargs(pdb_filepath: str, dqe_filepath: str) -> dict:
    """Get simulation keyword arguments for the test."""
    # Base parameters
    kwargs = {
        "beam_energy_kev": 300,
        "sim_pixel_spacing": 0.95,
        "sim_volume_shape": (400, 400, 400),
        "requested_upsampling": -1,
        "apply_dose_weighting": True,
        "dose_start": 0,
        "dose_end": 50,
        "dose_filter_modify_signal": "None",
        "dose_filter_critical_bfactor": -1.0,
        "apply_dqe": True,
    }

    # Load in atom data from pdb file
    atom_positions_zyx, atom_ids, atom_b_factors = load_model(pdb_filepath)
    atom_positions_zyx, atom_ids, atom_b_factors = remove_hydrogens(
        atom_positions_zyx, atom_ids, atom_b_factors
    )
    atom_b_factors *= 0.25  # NOTE: Scaling to match cisTEM b-factor scaling

    kwargs["atom_positions_zyx"] = atom_positions_zyx
    kwargs["atom_ids"] = atom_ids
    kwargs["atom_b_factors"] = atom_b_factors

    # Load in DQE data from star file
    mtf_frequencies, mtf_amplitudes = read_mtf(dqe_filepath)
    kwargs["mtf_frequencies"] = mtf_frequencies
    kwargs["mtf_amplitudes"] = mtf_amplitudes

    return kwargs


def is_ci() -> bool:
    """Check if running in CI environment."""
    return os.environ.get("CI", "false").lower() == "true"


@pytest.mark.skipif(
    is_ci(),
    reason="Skip on CI due to memory constraints. Run locally or on large runner only.",
)
def test_simulate3d():
    """Do the simulation and compare to the reference mrc file."""
    pdb_filepath, mrc_filepath, dqe_filepath = setup_simulation()

    # Ensure files are downloaded
    assert os.path.exists(pdb_filepath), f"File not found: {pdb_filepath}"
    assert os.path.exists(mrc_filepath), f"File not found: {mrc_filepath}"
    assert os.path.exists(dqe_filepath), f"File not found: {dqe_filepath}"

    simulate3d_kwargs = get_simulation_kwargs(pdb_filepath, dqe_filepath)

    # Run the simulation
    simulated_volume = simulate3d(**simulate3d_kwargs)
    simulated_volume = simulated_volume.cpu().numpy()

    # Compare the simulated volume to the reference mrc file
    with (
        mrcfile.open(mrc_filepath) as ref_mrc,
    ):
        assert simulated_volume.shape == ref_mrc.data.shape
        assert np.allclose(simulated_volume, ref_mrc.data, atol=1e-3)



"""For handling MRC files."""

import mrcfile
import numpy as np
import torch


def tensor_to_mrc(
    output_filename: str,
    final_volume: torch.Tensor,
    sim_pixel_spacing: float,
) -> None:
    """Write the final volume to an MRC file.

    NOTE: This function is only for 3D volumetric data; 2D data is not supported.

    Parameters
    ----------
    output_filename : str
        Path to the output MRC file.
    final_volume : torch.Tensor
        Volume information to write to the MRC file.
    sim_pixel_spacing :
        The pixel spacing in the simulation.

    Returns
    -------
    None
    """
    write_volume = final_volume.cpu().numpy()
    with mrcfile.new(output_filename, overwrite=True) as mrc:
        mrc.set_data(write_volume)
        mrc.voxel_size = (sim_pixel_spacing, sim_pixel_spacing, sim_pixel_spacing)
        # Populate more of the metadata...
        # Header setup
        mrc.header.mode = 2  # Mode 2 is float32
        mrc.header.nx, mrc.header.ny, mrc.header.nz = write_volume.shape[
            ::-1
        ]  # Dimensions (x, y, z)
        mrc.header.mx, mrc.header.my, mrc.header.mz = write_volume.shape[
            ::-1
        ]  # Sampling grid size (same as data dimensions)
        cell_dimensions = [dim * sim_pixel_spacing for dim in write_volume.shape[::-1]]
        mrc.header.cella.x = cell_dimensions[0]  # X dimension in angstroms
        mrc.header.cella.y = cell_dimensions[1]  # Y dimension in angstroms
        mrc.header.cella.z = cell_dimensions[2]  # Z dimension in angstroms

        mrc.header.mapc = 1  # Columns correspond to x
        mrc.header.mapr = 2  # Rows correspond to y
        mrc.header.maps = 3  # Sections correspond to z

        mrc.header.dmin = write_volume.min()  # Minimum density value
        mrc.header.dmax = write_volume.max()  # Maximum density value
        mrc.header.dmean = write_volume.mean()  # Mean density value

        # Additional metadata
        mrc.update_header_from_data()  # Automatically populates remaining fields
        mrc.header.rms = np.std(write_volume)  # RMS deviation of the density values



"""Calculates the scatttering potential."""

import json
from pathlib import Path

import einops
import torch


def get_scattering_parameters() -> tuple[dict, dict]:
    """Load scattering parameters from JSON file.

    NOTE: This function currently assumes the JSON file path relative to this
    file in the package. Could add future functionality to specify the path.

    Parameters
    ----------
    None

    Returns
    -------
    scattering_params_a : dict
        Scattering parameters for atom type A.
    scattering_params_b : dict
        Scattering parameters for atom type B.
    """
    scattering_param_path = Path(__file__).parent / "elastic_scattering_factors.json"

    with open(scattering_param_path) as f:
        data = json.load(f)

    scattering_params_a = {k: v for k, v in data["parameters_a"].items() if v != []}
    scattering_params_b = {k: v for k, v in data["parameters_b"].items() if v != []}

    return scattering_params_a, scattering_params_b


SCATTERING_PARAMS_A, SCATTERING_PARAMS_B = get_scattering_parameters()


def get_a_param(atom_ids: list[str]) -> torch.Tensor:
    """Get the 'a' scattering parameters for a set of atoms.

    Parameters
    ----------
    atom_ids : list[str]
        Atom IDs as a list of uppercase element symbols.

    Returns
    -------
    params_tensor: torch.Tensor
        Scattering parameters for each atom in the neighborhood
    """
    # Iterate over each atom_id in atoms_id_filtered
    params_list = [SCATTERING_PARAMS_A[atom_id] for atom_id in atom_ids]
    params_tensor = torch.tensor(params_list)

    return params_tensor


def get_b_param(atom_ids: list[str]) -> torch.Tensor:
    """Get the 'b' scattering parameters for a set of atoms.

    Parameters
    ----------
    atom_ids : list[str]
        Atom IDs as a list of uppercase element symbols.

    Returns
    -------
    params_tensor: torch.Tensor
        Scattering parameters for each atom in the neighborhood
    """
    # Iterate over each atom_id in atoms_id_filtered
    params_list = [SCATTERING_PARAMS_B[atom_id] for atom_id in atom_ids]
    params_tensor = torch.tensor(params_list)

    return params_tensor


def get_total_b_param(
    atom_ids: list[str],
    atom_b_factors: torch.Tensor,
) -> torch.Tensor:
    """Calculate the total B parameter per atom.

    Parameters
    ----------
    atom_ids : list[str]
        Atom IDs as a list of uppercase element symbols.
    atom_b_factors : torch.Tensor
        Atom B factors.

    Returns
    -------
    bPlusB: torch.Tensor
        Total B parameter for each atom in the neighborhood.
    """
    b_params = get_b_param(atom_ids)
    bPlusB = 2 * torch.pi / torch.sqrt(atom_b_factors.unsqueeze(1) + b_params)

    return bPlusB


def get_scattering_potential_of_voxel_batch(
    zyx_coords1: torch.Tensor,  # Shape: (atomN, voxelN, 3)
    zyx_coords2: torch.Tensor,  # Shape: (atomN, voxelN, 3)
    atom_ids: list[str],  # Shape: (atomN)
    atom_b_factors: torch.Tensor,  # Shape: (atomN)
    lead_term: float,
    device: torch.device = None,
) -> torch.Tensor:
    """Batched calculation over atoms for scattering potential at each voxel.

    Follows the equation 12 from https://journals.iucr.org/m/issues/2021/06/00/rq5007/index.html
    for calculating the scattering potential per voxel.

    Parameters
    ----------
    zyx_coords1 : torch.Tensor
        Coordinates of the first voxel in the neighborhood. Shape of
        (atomN, voxelN, 3) where atomN is the number of atoms and voxelN
        is the number of voxels in the neighborhood.
    zyx_coords2 : torch.Tensor
        Coordinates of the second voxel in the neighborhood. Shape of
        (atomN, voxelN, 3) where atomN is the number of atoms and voxelN
        is the number of voxels in the neighborhood.
    atom_ids : list[str]
        Atom IDs as a list of uppercase element symbols.
    atom_b_factors : torch.Tensor
        Atom B factors.
    lead_term : float
        Lead term for the scattering potential.
    device: torch.device
        Device to run the computation on.

    Returns
    -------
    potential : torch.Tensor
        Scattering potential for all voxels in the neighborhood.
    """
    # a   -> atom index with shape (n_atoms)
    # d   -> dimension with shape (3)
    # v/n -> flattened voxel indices for neighborhood with shape (h*w*d)
    # i   -> scattering fit index with shape (5)

    # If device not specified, use the device of input tensors
    if device is None:
        device = zyx_coords1.device

    # Get scattering parameters for atoms
    params_a = get_a_param(atom_ids)
    params_bPlusB = get_total_b_param(atom_ids, atom_b_factors)

    # Compare signs element-wise for batched coordinates
    t_all = (zyx_coords1 * zyx_coords2) >= 0  # Shape: (atomN, voxelN, 3)
    temp_potential = torch.zeros(len(zyx_coords1), device=device)

    # rearrange for broadcasting
    zyx_coords1 = einops.rearrange(zyx_coords1, "a n d -> a n d 1")
    zyx_coords2 = einops.rearrange(zyx_coords2, "a n d -> a n d 1")
    t_all = einops.rearrange(t_all, "a n d -> a n d 1")
    params_bPlusB = einops.rearrange(params_bPlusB, "a i -> a 1 1 i")

    all_terms = torch.where(
        t_all,
        torch.special.erf(params_bPlusB * zyx_coords2)
        - torch.special.erf(params_bPlusB * zyx_coords1),
        torch.abs(torch.special.erf(params_bPlusB * zyx_coords2))
        + torch.abs(torch.special.erf(params_bPlusB * zyx_coords1)),
    )
    t0 = einops.reduce(all_terms, "a n d i -> a n i", "prod")
    params_a = einops.rearrange(params_a, "a i -> a 1 i")
    a_mult = torch.abs(t0) * params_a
    temp_potential = einops.reduce(a_mult, "a n i -> a n", "sum")

    return (lead_term * temp_potential).float()



"""Handle PDB related operations."""

import pathlib

import mmdf
import torch


def load_model(
    file_path: str | pathlib.Path,
) -> tuple[torch.Tensor, list[str], torch.Tensor]:
    """Pdb file to atom coordinates, ids, and B factors.

    Loads a pdb from `file_path` and returns the atom coordinates (in Angstroms),
    atom ids as a list of strings, and B factors (in Angstroms^2).

    Parameters
    ----------
    file_path : str
        Path to the pdb file.

    Returns
    -------
    tuple[torch.Tensor, list[str], torch.Tensor]
        Atom coordinates, atom ids, and B factors.
    """
    df = mmdf.read(file_path)
    atom_zyx = torch.tensor(df[["z", "y", "x"]].to_numpy()).float()  # (n_atoms, 3)
    atom_zyx -= torch.mean(atom_zyx, dim=0, keepdim=True)  # center
    atom_id = df["element"].str.upper().tolist()
    atom_b_factor = torch.tensor(df["b_isotropic"].to_numpy()).float()

    return atom_zyx, atom_id, atom_b_factor


def remove_hydrogens(
    atoms_zyx: torch.Tensor,
    atoms_id: list,
    atoms_b_factor_scaled: torch.Tensor,
) -> tuple[torch.Tensor, list[str], torch.Tensor]:
    """
    Remove hydrogen atoms from the atom list.

    Parameters
    ----------
    atoms_zyx : torch.Tensor
        Atom coordinates in Angstroms.
    atoms_id : list[str]
        Atom IDs.
    atoms_b_factor_scaled : torch.Tensor
        Atom B factors.

    Returns
    -------
    tuple[torch.Tensor, list[str], torch.Tensor]
        Atom coordinates, atom IDs, and B factors with hydrogen atoms removed.
    """
    non_h_mask = [aid != "H" for aid in atoms_id]
    atoms_zyx_filtered = atoms_zyx[non_h_mask]
    atoms_id_filtered = [aid for i, aid in enumerate(atoms_id) if non_h_mask[i]]
    atoms_b_factor_scaled_filtered = atoms_b_factor_scaled[non_h_mask]
    return atoms_zyx_filtered, atoms_id_filtered, atoms_b_factor_scaled_filtered



import torch

from ttsim3d.scattering_potential import (
    get_a_param,
    get_b_param,
    get_total_b_param,
)


def test_get_scattering_parameters():
    """TODO: Construct this test"""
    pass


def test_get_a_param():
    """Test the get_a_param function."""
    atom_ids = ["H", "N", "O", "C"]

    # NOTE: These are just copied from the elastic_scattering_factors.json file
    expected = torch.tensor(
        [
            [0.0349, 0.1201, 0.1970, 0.0573, 0.1195],
            [0.1022, 0.3219, 0.7982, 0.8197, 0.1715],
            [0.0974, 0.2921, 0.6910, 0.6990, 0.2039],
            [0.0893, 0.2563, 0.7570, 1.0487, 0.3575],
        ]
    )
    result = get_a_param(atom_ids)

    assert isinstance(result, torch.Tensor), f"Expected a tensor, got {type(result)}"
    assert torch.allclose(
        result, expected
    ), "Mismatch between expected a parameters and result"


def test_get_b_param():
    """Test the get_b_param function."""
    atom_ids = ["H", "N", "O", "C"]

    # NOTE: These are just copied from the elastic_scattering_factors.json file
    expected = torch.tensor(
        [
            [0.5347, 3.5867, 12.3471, 18.9525, 38.6269],
            [0.2451, 1.7481, 6.1925, 17.3894, 48.1431],
            [0.2067, 1.3815, 4.6943, 12.7105, 32.4726],
            [0.2465, 1.7100, 6.4094, 18.6113, 50.2523],
        ]
    )

    result = get_b_param(atom_ids)

    assert isinstance(result, torch.Tensor), f"Expected a tensor, got {type(result)}"

    assert torch.allclose(
        result, expected
    ), "Mismatch between expected b parameters and result"


def test_get_total_b_param():
    """Test for combined b parameter and B-factor."""
    atom_ids = ["H", "N", "O", "C"]
    atom_b_factors = torch.tensor([40.0, 30.0, 20.0, 10.0])

    # expected is 2 * PI / sqrt(B_n + b)
    expected = torch.tensor(
        [
            [0.986885, 0.951706, 0.868428, 0.818331, 0.708589],
            [1.142490, 1.115119, 1.044409, 0.912724, 0.710779],
            [1.397759, 1.358816, 1.264391, 1.098592, 0.867388],
            [1.962873, 1.836122, 1.551078, 1.174657, 0.809456],
        ]
    )

    result = get_total_b_param(atom_ids, atom_b_factors)

    assert isinstance(result, torch.Tensor), f"Expected a tensor, got {type(result)}"

    assert torch.allclose(
        result, expected
    ), "Mismatch between expected total b parameters and result"



"""Simulate a 3D electrostatic potential map from a PDB in pyTorch."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("ttsim3d")
except PackageNotFoundError:
    __version__ = "uninstalled"
__author__ = "Josh Dickerson"
__email__ = "jdickerson@berkeley.edu"



"""Run tests for the Simulator and SimulatorConfig pydantic models."""

import pytest
from pydantic import ValidationError

from ttsim3d.models import DEFAULT_MTF_REFERENCES, SimulatorConfig


def good_simulator_config() -> dict:
    """Returns a dictionary of valid inputs for the SimulatorConfig model."""
    return {
        "voltage": 300,
        "apply_dose_weighting": True,
        "crit_exposure_bfactor": -1,
        "dose_filter_modify_signal": "None",
        "dose_start": 0.0,
        "dose_end": 30.0,
        "apply_dqe": True,
        "mtf_reference": "k2_300kv",
        "upsampling": -1,
        "store_volume": False,
    }


def test_default_mtf_references():
    """Ensure the default MTF references are valid."""
    good_dict = good_simulator_config()
    mtf_reference_strs = list(DEFAULT_MTF_REFERENCES.keys())

    # Ensure all default MTF references are valid
    for mtf_ref in mtf_reference_strs:
        good_dict["mtf_reference"] = mtf_ref
        _ = SimulatorConfig(**good_dict)


def test_simulator_config_validation():
    """Basic tests for the validation of SimulatorConfig model inputs."""
    good_dict = good_simulator_config()

    # First, check that a valid dictionary passes validation
    _ = SimulatorConfig(**good_dict)

    # Ensure voltage is positive
    bad_voltage = good_dict.copy()
    bad_voltage["voltage"] = -300
    with pytest.raises(ValidationError):
        SimulatorConfig(**bad_voltage)

    # Ensure the dose_filter_modify_signal is a valid choice
    bad_modify_signal = good_dict.copy()
    bad_modify_signal["dose_filter_modify_signal"] = "invalid_choice"
    with pytest.raises(ValidationError):
        SimulatorConfig(**bad_modify_signal)

    # Ensure invalid mtf paths raise an error
    bad_mtf = good_dict.copy()
    bad_mtf["mtf_reference"] = "/some/invalid/path/mtf.star"
    with pytest.raises(ValidationError):
        SimulatorConfig(**bad_mtf)


def test_simulator_validation():
    """Basic tests for the validation of Simulator model inputs."""
    # TODO: Implement test
    pass



"""Handles cpu/gpu device selection."""

from typing import Optional

import torch


def select_gpu(
    gpu_id: Optional[int] = None,
) -> torch.device:
    """
    Select multiple GPU devices based on IDs or available memory.

    Args:
        gpu_id: Specific GPU ID to use.
                If None, selects GPU with most available memory.

    Returns
    -------
        torch.device: Selected GPU device or CPU if no GPU available
    """
    # Check if you can actually use a cuda gpu
    if not torch.cuda.is_available():
        print("No GPU available, using CPU")
        return [torch.device("cpu")]

    n_gpus = torch.cuda.device_count()
    if n_gpus == 0:
        print("No GPU available, using CPU")
        return [torch.device("cpu")]

    # If specific GPUs requested, validate and return them
    if gpu_id is not None:
        valid_device = None
        if gpu_id >= n_gpus:
            print(f"Requested GPU {gpu_id} not available. Max GPU ID is {n_gpus-1}")
        else:
            valid_device = torch.device(f"cuda:{gpu_id}")

        if valid_device is None:
            print("No valid GPUs specified. Using CPU")
            return [torch.device("cpu")]
        return valid_device

    # Find GPU with most available memory
    gpu_memory_available = []
    print("\nAvailable GPUs:")
    for i in range(n_gpus):
        torch.cuda.set_device(i)
        total_memory = torch.cuda.get_device_properties(i).total_memory
        allocated_memory = torch.cuda.memory_allocated(i)
        available = total_memory - allocated_memory

        print(f"GPU {i}: {torch.cuda.get_device_properties(i).name}")
        print(f"  Total memory: {total_memory/1024**3:.1f} GB")
        print(f"  Available memory: {available/1024**3:.1f} GB")

        gpu_memory_available.append((i, available))

    # Select the GPU with the most available memory
    selected_gpu_idx, _ = max(gpu_memory_available, key=lambda x: x[1])
    selected_gpu = torch.device(f"cuda:{selected_gpu_idx}")

    print("\nSelected GPU:", selected_gpu)
    return selected_gpu



"""Simple run script."""

import click

from ttsim3d.models import Simulator, SimulatorConfig


@click.command()
@click.option(
    "--pdb-filepath",
    type=click.Path(),
    required=True,
    help="The path to the PDB file containing the atomic structure to simulate.",
)
@click.option(
    "--mrc-filepath",
    type=click.Path(),
    required=True,
    help="File path to save simulated volume.",
)
@click.option(
    "--pixel-spacing",
    type=float,
    help=(
        "The pixel spacing of the simulated volume in units of Angstroms. "
        "Must be greater than 0."
    ),
)
@click.option(
    "--volume-shape",
    type=(int, int, int),
    help="The shape of the simulated volume in pixels.",
)
@click.option(
    "--voltage",
    type=float,
    default=300.0,
    help="The voltage of the microscope in kV. Default is 300 kV.",
)
@click.option(
    "--upsampling",
    type=int,
    default=-1,
    help=(
        "The upsampling factor to apply to the simulation. The default is -1 "
        "and corresponds to automatic calculation of the upsampling factor."
    ),
)
@click.option(
    "--b-factor-scaling",
    type=float,
    default=1.0,
    help=(
        "The scaling factor to apply to the B-factors of the atoms in the "
        "pdb file. The default is 1.0."
    ),
)
@click.option(
    "--additional-b-factor",
    type=float,
    default=0.0,
    help=(
        "Additional B-factor to apply to the atoms in the pdb file. The "
        "default is 0.0."
    ),
)
@click.option(
    "--apply-dose-weighting",
    type=bool,
    default=True,
    help="If True, apply dose weighting to the simulation. Default is True.",
)
@click.option(
    "--crit-exposure-bfactor",
    type=float,
    default=-1.0,
    help=(
        "B-factor to use in critical exposure calculations. The default is -1 "
        "and corresponds to the fitted critical exposure function in Grant "
        "and Grigorieff, 2015."
    ),
)
@click.option(
    "--dose-filter-modify-signal",
    type=click.Choice(["None", "sqrt", "rel_diff"]),
    default="None",
    help=(
        "Signal modification to apply to the dose filter. Currently supports "
        "'None', 'sqrt', and 'rel_diff'."
    ),
)
@click.option(
    "--dose-start",
    type=float,
    default=0.0,
    help="The starting dose in e/A^2. Default is 0.0.",
)
@click.option(
    "--dose-end",
    type=float,
    default=30.0,
    help="The ending dose in e/A^2. Default is 30.0.",
)
@click.option(
    "--apply-dqe",
    is_flag=True,
    help="If True, apply a DQE filter to the simulation.",
)
@click.option(
    "--mtf-reference",
    type=str,
    default="k2_300kV",
    help=(
        "Path to the modulation transfer function (MTF) reference star "
        "file, or one of the known MTF reference files. "
        "Default is 'k2_300kV'."
    ),
)
@click.option(
    "--gpu-ids",
    type=list[int],
    multiple=True,
    help="A list of GPU IDs to use for the simulation. Currently unused.",
)
def main(
    pdb_filepath: str,
    mrc_filepath: str,
    pixel_spacing: float,
    volume_shape: tuple[int, int, int],
    voltage: float,
    upsampling: int,
    b_factor_scaling: float,
    additional_b_factor: float,
    apply_dose_weighting: bool,
    crit_exposure_bfactor: float,
    dose_filter_modify_signal: str,
    dose_start: float,
    dose_end: float,
    apply_dqe: bool,
    mtf_reference: str,
    gpu_ids: list[int],
) -> None:
    """Run a structure simulation through the CLI."""
    simulator_config = SimulatorConfig(
        voltage=voltage,
        apply_dose_weighting=apply_dose_weighting,
        crit_exposure_bfactor=crit_exposure_bfactor,
        dose_filter_modify_signal=dose_filter_modify_signal,  # type: ignore
        dose_start=dose_start,
        dose_end=dose_end,
        apply_dqe=apply_dqe,
        mtf_reference=mtf_reference,
        upsampling=upsampling,
    )
    simulator = Simulator(
        pixel_spacing=pixel_spacing,
        volume_shape=volume_shape,
        pdb_filepath=pdb_filepath,
        b_factor_scaling=b_factor_scaling,
        additional_b_factor=additional_b_factor,
        simulator_config=simulator_config,
    )
    simulator.export_to_mrc(mrc_filepath=mrc_filepath)


if __name__ == "__main__":
    main()
